import json
from typing import Dict

from celery import shared_task
import requests
from .open_observe_config import OpenObserveConfig
from .logged_classes import (InfoLogData, ExceptionLogData,
ResponseLogData, LogLevel
                             )
from django.http import HttpRequest, HttpResponse


import logging
from requests.auth import HTTPBasicAuth




class OpenObserveSender:
    def __init__(self, config : OpenObserveConfig) -> None:
        self.auth = HTTPBasicAuth(config.username, config.password)  # Store auth credentials
        open_obseve_url = f"{config.open_observe_host}/api/{config.open_observe_organization_name}/{config.open_observe_stream_name}/_json"
        self.open_observe_url = open_obseve_url
        self.logger = logging.getLogger("RequestLogger")
        self.logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

    def log_response(self, request:HttpRequest,response:HttpResponse, duration_in_ms:float ) -> None:
        response_log_data = ResponseLogData(request=request,response=response,duration_ms=duration_in_ms)
        # Send log to external log server
        self.send_log_to_open_observe(response_log_data.get_log_data())

    def log_exception(self, exception: Exception, request: HttpRequest) -> None:
        """
        Log unhandled exceptions.
        
        :param exception: The exception object.
        """
        exception_log = ExceptionLogData(exception=exception, request=request)
        # Send log to external log server
        self.send_log_to_open_observe(exception_log.get_log_data())

    def send_log_to_open_observe(self, log_data: dict) -> None:
        """
        Private method to send log data to a remote server.
        
        :param log_data: Log data in JSON format.
        """
        send_log_async_with_celery.delay( open_observe_url= self.open_observe_url, log_data=log_data, username=self.auth.username, password=self.auth.password)
            

@shared_task
def send_log_async_with_celery(open_observe_url:str,username:str, password:str, log_data:dict):
    try:
        headers = {'Content-Type': 'application/json'}
        auth = HTTPBasicAuth(username=username, password=password)
        r = requests.post(open_observe_url, auth=auth, data=json.dumps(log_data), headers=headers)
        if r.status_code != 200:
          logging.error(f"Failed to send log to server: {r.text}")
        else:
          logging.info(f"Log sent to openobserve")

    except requests.exceptions.RequestException as e:
          logging.error(f"Failed to send log to server: {r.text}")


  
  





# The sender has to be setted up with setup_open_observe_sender function
from .open_observe_config import OpenObserveConfig
from .open_observe_sender import OpenObserveSender

from django.conf import settings

def setup_open_observe_sender()->OpenObserveSender:
  config_from_settings = settings.OPENOBSERVE_CONFIG
  config = OpenObserveConfig(
    username=config_from_settings["username"],
    password=config_from_settings["password"],
    open_observe_host=config_from_settings["open_observe_host"],
    open_observe_organization_name=config_from_settings["open_observe_organization_name"],
    open_observe_stream_name=config_from_settings["open_observe_stream_name"],
  )
  new_sender = OpenObserveSender(config=config)
  return new_sender


class OpenObserveConfig:

    def _clean_string(self,value: str) -> str:
      """
      Removes start and ending whitespace and start and ending '/'
      """
      if value is None:
          raise ValueError("The parameter cannot be null.")
      cleaned_value = value.strip().strip('/')
      if not cleaned_value:
          raise ValueError("The parameter cannot be empty after trimming.")
      return cleaned_value

    def __init__(
        self, 
        open_observe_host: str, 
        username: str, 
        password: str, 
        open_observe_organization_name: str, 
        open_observe_stream_name: str
    ) -> None:
        # Validate parameters
        if not open_observe_host:
            raise ValueError("The 'open_observe_host' parameter cannot be null or empty.")
        if not username:
            raise ValueError("The 'username' parameter cannot be null or empty.")
        if not password:
            raise ValueError("The 'password' parameter cannot be null or empty.")
        if not open_observe_organization_name:
            raise ValueError("The 'open_observe_organization_name' parameter cannot be null or empty.")
        if not open_observe_stream_name:
            raise ValueError("The 'open_observe_stream_name' parameter cannot be null or empty.")
        
        # Store parameters
        self.open_observe_host = self._clean_string(open_observe_host)
        self.username = self._clean_string(username)
        self.password = self._clean_string(password)
        self.open_observe_organization_name = self._clean_string(open_observe_organization_name)
        self.open_observe_stream_name = self._clean_string(open_observe_stream_name)

    #function to print the config
    def __repr__(self) -> str:
        return (f"OpenObserveConfig("
                f"open_observe_host='{self.open_observe_host}', "
                f"username='{self.username}', "
                f"password='{self.password}', "
                f"open_observe_organization_name='{self.open_observe_organization_name}', "
                f"open_observe_stream_name='{self.open_observe_stream_name}')")





import time
from typing import Callable

from django.http import HttpRequest, HttpResponse
from .sender_creator import setup_open_observe_sender


class OpenObserveTrackingMiddleware:
    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        self.get_response = get_response
        self.open_openobserve_sender = setup_open_observe_sender()

    def process_exception(self, request, exception):
        """
        This captures an exception from the self.get_response() inside the __call__ method
        """
        self.open_openobserve_sender.log_exception(exception=exception,request=request)  # Log the exception
        self.had_unhandled_exception=True
        raise exception # re raise exception if you want another middleware to handle the request with custom messages

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """
        This captures the request before it gets to the view and after the view creates a response
        """
        # Before sending the request to the view
        self.had_unhandled_exception = False
        start_time = time.perf_counter()
        # Process the request and get the response
        response = self.get_response(request)  # Can throw and exception if there was an unhandled exception on the view
        # After the view finish processing
        duration_in_ms = (time.perf_counter()-start_time) *1000
        duration_formatted = round(duration_in_ms, 4)
        print(duration_formatted)
        # If no expection Log the request and response details
        if self.had_unhandled_exception == False:
          self.open_openobserve_sender.log_response(  request=request, response=response,duration_in_ms=duration_formatted)
        return response


from enum import Enum
from http import HTTPStatus
import traceback
from typing import Dict
import uuid
from django.http import HttpRequest, HttpResponse


STATUS_CODES_MAP = {status.value: status.phrase for status in HTTPStatus}

class ExceptionLogData:
  def __init__(self,exception: Exception, request: HttpRequest):
    self.url = request.get_full_path(force_append_slash=True)
    self.status_code = 500
    self.method = request.method
    self.exception = exception
    self.traceback = traceback.format_exc()
  def get_log_data(self)->Dict:
    log_data = {
        "url":self.url,
        "status_code": self.status_code,
        "method":self.method,
        "exception": str(self.exception),
        "traceback": str(self.traceback),
        "is_automatic_log": True
    }
    return log_data


class LogLevel(Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

class InfoLogData:
  def __init__(self, request: HttpRequest, log_info:Dict ,log_level : LogLevel):
    self.url = request.get_full_path(force_append_slash=True)
    self.method = request.method
    self.log_info= log_info
    if not hasattr(request,'open_observe_id'):
      open_observe_id = str(uuid.uuid4())
      request.open_observe_id = open_observe_id
    self.open_observe_id = request.open_observe_id
    self.tag = log_level

  def get_log_data(self)->Dict:
    log_data = {
        "url":self.url,
        "method":self.method,
        "log_level": self.tag._value_,
        "open_observe_request_id": self.open_observe_id,
        "info": self.log_info,
        "is_automatic_log": False,
    }

    return log_data


class ResponseLogData:
  def __init__(self, request: HttpRequest, response: HttpResponse, duration_ms: float):
    self.url = request.get_full_path(force_append_slash=True)
    self.method = request.method
    self.status_code = response.status_code
    self.duration_in_ms = duration_ms
    self.error_message = None

    if response.status_code >400 and response.status_code < 599 : #if status is client error or server error then log the error message
      response_body = None
      content_type = response.headers.get('Content-Type', '')
      #Add body to log if it's not html 
      if 'text/html'  in content_type:
          response_body = "Body is HTML, not logged to optimize resources"
      else:
          try:
            response_body = response.content.decode('utf-8')
          except UnicodeDecodeError:
            response_body = "<binary data>"
      self.error_message = response_body
  def get_log_data(self)->Dict:
    log_data = {
        "method": self.method,
        "url": self.url,
        "status_code": self.status_code,
        "duration_in_ms": self.duration_in_ms,
        "status_code_words": STATUS_CODES_MAP.get(self.status_code),
        "error_message": self.error_message,
        "is_automatic_log": True
    }
    return log_data


from setuptools import setup, find_packages


with open("README.md", "r") as f:
  description = f.read()


setup(
  name='open_observe_tracking_django',
  version='0.2',
  packages= find_packages(),
  install_requires=[
    'requests',
    'Django>=4.1.6',
    'celery',

  ],
  long_description=description,
  long_description_content_type="text/markdown"
  

)



from typing import Dict
from django.http import HttpRequest

from open_observe_tracking_django.logged_classes import InfoLogData, LogLevel

from .sender_singletone import setup_open_observe_sender

sender = None
def log_to_open_observe(request:HttpRequest,log_level : LogLevel,info:Dict ):
  """
  Send logs to open observe and to the terminal

  :param request: To identify the request in which the log was sent
  :param log_level: The tag you want to add to the log level
  :param info: A dictionary to be sent to open observe, ("key": value pairs)

  ## Example:

    info = {"user_id": 32323, "user_name": "Than", "message": "Is trying to send messages to blocked user"}
    # Call the logging function
    log_to_open_observe(request, LogLevel.WARNING, info)
    
    # Another
    info = {"error","Database is not responding"}
    # Call the logging function
    log_to_open_observe(request, LogLevel.CRITICAL, info)
  """
  global sender
  if sender == None:
    sender = setup_open_observe_sender()
    print("CREANDO")

  log = InfoLogData(request=request,log_info=info,log_level=log_level)
  sender.send_log_to_open_observe(log_data=log.get_log_data())